"""
Interfaces for interacting with "dependency sources", i.e. sources
of fully resolved Python dependency trees.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterator

from packaging.requirements import Requirement

from pip_audit._fix import ResolvedFixVersion
from pip_audit._service import Dependency


class DependencySource(ABC):
    """
    Represents an abstract source of fully-resolved Python dependencies.

    Individual concrete dependency sources (e.g. `pip list`) are expected
    to subclass `DependencySource` and implement it in their terms.
    """

    @abstractmethod
    def collect(self) -> Iterator[Dependency]:  # pragma: no cover
        """
        Yield the dependencies in this source.
        """
        raise NotImplementedError

    @abstractmethod
    def fix(self, fix_version: ResolvedFixVersion) -> None:  # pragma: no cover
        """
        Upgrade a dependency to the given fix version.
        """
        raise NotImplementedError


class DependencySourceError(Exception):
    """
    Raised when a `DependencySource` fails to provide its dependencies.

    Concrete implementations are expected to subclass this exception to
    provide more context.
    """

    pass


class DependencyFixError(Exception):
    """
    Raised when a `DependencySource` fails to perform a "fix" operation, i.e.
    fails to upgrade a package to a different version.

    Concrete implementations are expected to subclass this exception to provide
    more context.
    """

    pass


class HashMissingError(Exception):
    """
    Raised when `RequirementHashes` fails to find any hashes for a resolved
    dependency.
    """

    pass


class HashMismatchError(Exception):
    """
    Raised when `RequirementHashes` fails to match a hash for a given
    requirement.
    """

    pass


class UnsupportedHashAlgorithm(Exception):
    """
    Raised when a `DependencyResolver` encounters an unsupported hash algorithm
    in the supplied `RequirementHashes`.
    """

    pass


class RequirementHashes:
    """
    Represents the hashes contained within a requirements file.
    """

    def __init__(self) -> None:
        """
        Create a new `RequirementHashes`.
        """
        self.mapping: dict[str, dict[str, list[str]]] = {}

    def add_req(self, req_name: str, hash_options_mapping: dict[str, list[str]]) -> None:
        """
        Add a set of hashes for a given requirement.

        `req_name`is the name of the requirement to check.

        `hash_options_mapping` is a dictionary mapping from algorithm names to a list of potential
        hashes. Requirements files are allowed to specify multiple hashes of the same algorithm to
        account for different distribution types.
        """
        self.mapping[req_name] = hash_options_mapping

    def __bool__(self) -> bool:
        """
        Check whether any requirements have been added.
        """
        return bool(self.mapping)

    def __contains__(self, req_name: str) -> bool:
        """
        Check whether a given requirement exists in the set of hashes.

        `req_name` is the name of the requirement to check.
        """
        return req_name in self.mapping

    def match(self, req_name: str, dist_hashes: dict[str, str]) -> None:
        """
        Check whether any of the provided hashes match the hashes calculated by the dependency
        resolver.

        `req_name` is the name of the requirement to check.

        `dist_hashes` is a mapping of hash algorithms to calculated hashes.
        """
        if req_name not in self.mapping:
            raise HashMissingError(f"No hashes found for {req_name}")

        for algorithm, hashes in self.mapping[req_name].items():
            for hash_ in hashes:
                if hash_ == dist_hashes[algorithm]:
                    return
        raise HashMismatchError(
            f"Mismatching hash for {req_name}, none of the calculated hashes ({dist_hashes}) "
            f"matched expected ({self.mapping[req_name]})"
        )

    def supported_algorithms(self, req_name: str) -> list[str]:
        """
        Returns a list of hash algorithms that are supported for a given requirement.

        `req_name` is the name of the requirement to check.
        """
        if req_name not in self.mapping:
            return []
        return list(self.mapping[req_name].keys())


class DependencyResolver(ABC):
    """
    Represents an abstract resolver of Python dependencies that takes a list of
    dependencies and returns all of their transitive dependencies.

    Concrete dependency sources may use a resolver as part of their
    implementation.
    """

    @abstractmethod
    def resolve(
        self, reqs: list[Requirement], req_hashes: RequirementHashes
    ) -> list[Dependency]:  # pragma: no cover
        """
        Resolve a list of `Requirement`s into a list of resolved `Dependency`s.
        """
        raise NotImplementedError


class DependencyResolverError(Exception):
    """
    Raised when a `DependencyResolver` fails to resolve its dependencies.

    Concrete implementations are expected to subclass this exception to
    provide more context.
    """

    pass



"""
A `resolvelib` provider implementation that resolves against PyPI.

Closely adapted from `resolvelib`'s examples, which are copyrighted by the `resolvelib`
authors under the ISC license.
"""

from __future__ import annotations

import hashlib
import itertools
import logging
from abc import ABC
from dataclasses import dataclass
from email.message import EmailMessage, Message
from email.parser import BytesParser
from io import BytesIO
from operator import attrgetter
from pathlib import Path
from subprocess import CalledProcessError
from tempfile import TemporaryDirectory
from typing import Any, BinaryIO, Iterator, Mapping, cast
from urllib.parse import urljoin, urlparse
from zipfile import ZipFile

import html5lib
import requests
from cachecontrol import CacheControl
from packaging.requirements import Requirement
from packaging.specifiers import InvalidSpecifier, SpecifierSet
from packaging.utils import canonicalize_name, parse_sdist_filename, parse_wheel_filename
from packaging.version import Version
from resolvelib.providers import AbstractProvider
from resolvelib.resolvers import RequirementInformation

from pip_audit._dependency_source import RequirementHashes, UnsupportedHashAlgorithm
from pip_audit._service import SkippedDependency
from pip_audit._state import AuditState
from pip_audit._util import python_version
from pip_audit._virtual_env import VirtualEnv, VirtualEnvError

logger = logging.getLogger(__name__)

# TODO: Final[Version] when our minimal Python is 3.8.
PYTHON_VERSION: Version = python_version()


@dataclass
class Candidate(ABC):
    """
    Represents a dependency candidate.
    """

    name: str


class ResolvedCandidate(Candidate):
    """
    Represents a resolved dependency candidate that has been retrieved from PyPI. A dependency being
    resolved may have multiple candidates, which go through a selection process guided by various
    weights (version, `sdist` vs. `wheel`, etc.)
    """

    def __init__(
        self,
        name: str,
        project: str,
        filename: Path,
        version: Version,
        *,
        url: str,
        extras: set[str],
        is_wheel: bool,
        reqs: list[Requirement],
        session: CacheControl,
        req_hashes: RequirementHashes,
        timeout: int | None = None,
        state: AuditState = AuditState(),
    ) -> None:
        """
        Create a new `Candidate`.
        """

        self.name = canonicalize_name(name)
        self.project = project
        self.filename = filename
        self.version = version
        self.url = url
        self.extras = extras
        self.is_wheel = is_wheel
        self.reqs = reqs
        self._session = session
        self.req_hashes = req_hashes
        self._timeout = timeout
        self._state = state

        self._metadata: Message | None = None
        self._dependencies: list[Requirement] | None = None
        self.dist_hashes: dict[str, str] = {}

    def __repr__(self) -> str:  # pragma: no cover
        """
        A string representation for `Candidate`.
        """
        if not self.extras:
            return f"<{self.name}=={self.version} wheel={self.is_wheel}>"
        return f"<{self.name}[{','.join(self.extras)}]=={self.version} wheel={self.is_wheel}>"

    @property
    def metadata(self) -> Message:
        """
        Return the package metadata for this candidate.
        """

        if self._metadata is None:
            self._state.update_state(f"Fetching metadata for {self.name} ({self.version})")

            response = self._session.get(self.url, timeout=self._timeout)
            response.raise_for_status()
            candidate_data = response.content
            self._generate_dist_hashes(candidate_data)

            if self.is_wheel:
                self._metadata = self._get_metadata_for_wheel(candidate_data)
            else:
                self._metadata = self._get_metadata_for_sdist(candidate_data)
        return self._metadata

    def _get_dependencies(self) -> Iterator[Requirement]:
        """
        Computes the dependency set for this candidate.
        """
        deps: list[str] = self.metadata.get_all("Requires-Dist", [])
        extras = self.extras if self.extras else [""]

        for d in deps:
            r = Requirement(d)
            if r.marker is None:
                yield r
            else:
                for e in extras:
                    if r.marker.evaluate({"extra": e}):
                        yield r  # pragma: no cover

    @property
    def dependencies(self) -> list[Requirement]:
        """
        Returns the list of `Requirement`s for this candidate.
        """
        if self._dependencies is None:
            self._dependencies = list(self._get_dependencies())
        return self._dependencies

    def _get_metadata_for_wheel(self, wheel_data: bytes) -> Message:
        """
        Extracts the metadata for this candidate, if it's a wheel.
        """
        self._state.update_state(f"Extracting wheel for {self.name} ({self.version})")

        with ZipFile(BytesIO(wheel_data)) as z:
            for n in z.namelist():
                if n.endswith(".dist-info/METADATA"):
                    p = BytesParser()
                    # NOTE: MyPy bug? ZipFile.open is treated as IO[bytes], which
                    # should be unified with BinaryIO but isn't.
                    return p.parse(cast(BinaryIO, z.open(n)), headersonly=True)

        # If we didn't find the metadata, return an empty dict
        return EmailMessage()  # pragma: no cover

    def _get_metadata_for_sdist(self, sdist_data: bytes) -> Message:
        """
        Extracts the metadata for this candidate, if it's a source distribution.
        """

        metadata = EmailMessage()
        self._generate_dist_hashes(sdist_data)

        with TemporaryDirectory() as pkg_dir:
            sdist = Path(pkg_dir) / self.filename.name
            sdist.write_bytes(sdist_data)

            self._state.update_state(
                f"Installing source distribution in isolated environment for {self.name} "
                f"({self.version})"
            )

            with TemporaryDirectory() as ve_dir:
                try:
                    ve = VirtualEnv([str(sdist)], self._state)
                    ve.create(ve_dir)
                except CalledProcessError as e:
                    # NOTE: Virtual environment creation failure with this error likely indicates
                    # some kind of layering violation, such as an incorrectly configured
                    # system Python interfering with a virtual environment's `python` or `pip`.
                    # We don't really have a cleaner place to capture and transform it,
                    # since it happens directly in the core `venv` standard library module.
                    # See: https://bugs.python.org/issue38705
                    # See: https://github.com/pypa/build/issues/294
                    raise VirtualEnvError("virtual environment creation failed internally") from e

                self._state.update_state(
                    f"Querying installed packages for {self.name} ({self.version})"
                )

                installed_packages = ve.installed_packages
                for name, version in installed_packages:
                    metadata["Requires-Dist"] = f"{name}=={str(version)}"

        return metadata

    def _generate_dist_hashes(self, dist_data: bytes) -> None:
        """
        Populate a mapping of hash algorithm names to hashes for the candidate.
        """
        if self.project not in self.req_hashes:
            return
        hash_algorithms = self.req_hashes.supported_algorithms(self.project)
        dist_hashes = {}
        for algorithm in hash_algorithms:
            if algorithm not in hashlib.algorithms_available:
                raise UnsupportedHashAlgorithm(
                    f"encountered hash with unknown algorithm: {algorithm}"
                )
            if algorithm not in hashlib.algorithms_guaranteed:  # pragma: no cover
                raise UnsupportedHashAlgorithm(
                    f"encountered hash with known but non-guaranteed algorithm: {algorithm}, this "
                    "won't necessarily work on other platforms"
                )
            hasher = hashlib.new(algorithm, dist_data)
            dist_hashes[algorithm] = hasher.hexdigest()
        self.dist_hashes = dist_hashes


@dataclass
class SkippedCandidate(Candidate):
    """
    Represents a skipped dependency candidate. When a project can't be found on PyPI, we propagate a
    `SkippedCandidate` to satisfy the dependency resolver and to signal to the caller that the
    candidate couldn't be resolved properly.
    """

    skip_reason: str


def get_project_from_indexes(
    index_urls: list[str],
    session: CacheControl,
    project: str,
    reqs: list[Requirement],
    extras: set[str],
    req_hashes: RequirementHashes,
    timeout: int | None,
    state: AuditState,
) -> Iterator[ResolvedCandidate]:
    """Return candidates from all indexes created from the project name and extras."""
    project_found = False
    for index_url in index_urls:
        # Not all indexes are guaranteed to have the project so this isn't an error
        # We should only return an error if it can't be found on ANY of the supplied index URLs
        try:
            yield from get_project_from_index(
                index_url, session, project, reqs, extras, req_hashes, timeout, state
            )
            project_found = True
        except PyPINotFoundError:
            pass
    if not project_found:
        raise PyPINotFoundError(
            f'Could not find project "{project}" on any of the supplied index URLs: {index_urls}'
        )


def get_project_from_index(
    index_url: str,
    session: CacheControl,
    project: str,
    reqs: list[Requirement],
    extras: set[str],
    req_hashes: RequirementHashes,
    timeout: int | None,
    state: AuditState,
) -> Iterator[ResolvedCandidate]:
    """Return candidates from an index created from the project name and extras."""

    # NOTE: The trailing slash is important here: without it, the `urljoin`
    # below will treat the final component as a file and strip during the join.
    # It's also strictly more correct, per PEP 503.
    project_url = urljoin(index_url, f"{project}/")
    response: requests.Response = session.get(project_url, timeout=timeout)
    if response.status_code == 404:
        raise PyPINotFoundError
    response.raise_for_status()
    data = response.content
    doc = html5lib.parse(data, namespaceHTMLElements=False)
    links = doc.findall(".//a")
    if not links:
        raise PyPINotFoundError
    for i in links:
        parsed_dist_url = urlparse(i.attrib["href"])

        # Per PEP 503: The distribution's URL can be relative, in which case
        # it's relative to the project's simple index URL.
        if not parsed_dist_url.netloc:
            dist_url = urljoin(project_url, parsed_dist_url.geturl())
        else:
            dist_url = parsed_dist_url.geturl()

        py_req = i.attrib.get("data-requires-python")
        # Skip items that need a different Python version
        if py_req:
            try:
                # NOTE: Starting with packaging==22.0, specifier parsing is
                # stricter: specifier components can only use the wildcard
                # comparison syntax on exact comparison operators (== and !=),
                # not on ordered operators like `>=`. There are existing
                # packages that use the invalid syntax in their metadata
                # however (like nltk==3.6, which does requires-python >= 3.5.*),
                # so we follow pip`'s behavior and ignore these specifiers.
                spec = SpecifierSet(py_req)
                if PYTHON_VERSION not in spec:
                    continue
            except InvalidSpecifier:
                logger.warning(f"invalid specifier set for Python version: {py_req}")

        path = parsed_dist_url.path
        filename = path.rpartition("/")[-1]

        # Handle wheels and source distributions
        try:
            if filename.endswith(".whl"):
                (name, version, _, _) = parse_wheel_filename(filename)
                is_wheel = True
            else:
                # If it doesn't look like a wheel, try to parse it as an
                # sdist. This will raise for incorrect looking filenames,
                # which we'll then skip via the exception handler.
                (name, version) = parse_sdist_filename(filename)
                is_wheel = False

            # TODO: Handle compatibility tags?
            yield ResolvedCandidate(
                name,
                project,
                Path(filename),
                version,
                url=dist_url,
                extras=extras,
                is_wheel=is_wheel,
                reqs=reqs,
                timeout=timeout,
                state=state,
                session=session,
                req_hashes=req_hashes,
            )
        except Exception:
            continue


class PyPIProvider(AbstractProvider):
    """
    An implementation of `resolvelib`'s `AbstractProvider` that uses
    the official Python Package Index.
    """

    def __init__(
        self,
        index_urls: list[str],
        req_hashes: RequirementHashes,
        session: requests.Session,
        timeout: int | None = None,
        state: AuditState = AuditState(),
    ):
        """
        Create a new `PyPIProvider`.

        `index_urls` is a list of package index URLs.

        `req_hashes` is a `RequirementHashes` to control what package/algorithm combinations we
        generate hashes for.

        `timeout` is an optional argument to control how many seconds the component should wait for
        responses to network requests.

        `cache_dir` is an optional argument to override the default HTTP caching directory.

        `state` is an `AuditState` to use for state callbacks.
        """

        # Per PEP 503: Index URLs should always be normalized to end with `/`.
        index_urls = [url if url.endswith("/") else f"{url}/" for url in index_urls]

        self.index_urls = index_urls
        self.req_hashes = req_hashes
        self.session = session
        self.timeout = timeout
        self._state = state
        self.skip_deps: list[SkippedDependency] = []

    def identify(self, requirement_or_candidate: Requirement | Candidate) -> str:
        """
        See `resolvelib.providers.AbstractProvider.identify`.
        """
        return canonicalize_name(requirement_or_candidate.name)

    # TODO: Typing. See: https://github.com/sarugaku/resolvelib/issues/104
    def get_preference(  # type: ignore[no-untyped-def]
        self,
        identifier: Any,
        resolutions: Mapping[Any, Any],
        candidates: Mapping[Any, Iterator[Any]],
        information: Mapping[Any, Iterator[RequirementInformation]],
        backtrack_causes: Any,
    ):
        """
        See `resolvelib.providers.AbstractProvider.get_preference`.
        """
        return sum(1 for _ in candidates[identifier])

    def find_matches(
        self,
        identifier: Any,
        requirements: Mapping[Any, Iterator[Any]],
        incompatibilities: Mapping[Any, Iterator[Any]],
    ) -> Iterator[Any]:
        """
        See `resolvelib.providers.AbstractProvider.find_matches`.
        """
        self._state.update_state(f"Resolving {identifier}")

        requirements = list(requirements[identifier])
        logger.debug(
            f"{identifier} req specifier constraints: {[r.specifier for r in requirements]}"
        )

        bad_versions = {c.version for c in incompatibilities[identifier]}

        # Accumulate extras
        extras: set[str] = set()
        for r in requirements:
            extras |= r.extras

        # Need to pass the extras to the search, so they
        # are added to the candidate at creation - we
        # treat candidates as immutable once created.
        all_candidates = get_project_from_indexes(
            self.index_urls,
            self.session,
            identifier,
            requirements,
            extras,
            self.req_hashes,
            self.timeout,
            self._state,
        )

        candidates: list[ResolvedCandidate]
        try:
            candidates = sorted(
                [
                    candidate
                    for candidate in all_candidates
                    if candidate.version not in bad_versions
                    # NOTE(ww): We use `filter(...)` instead of checking
                    # `candidate.version in r.specifier` because the former has subtle (and PEP 440
                    # mandated) behavior around prereleases. Specifically, `filter(...)`
                    # returns prereleases even if not explicitly configured, but only if
                    # there are no non-prereleases.
                    # See: https://github.com/pypa/pip-audit/issues/472
                    and all([any(r.specifier.filter((candidate.version,))) for r in requirements])
                    # HACK(ww): Additionally check that each candidate's name matches the
                    # expected project name (identifier).
                    # This technically shouldn't be required, but parsing distribution names
                    # from package indices is imprecise/unreliable when distribution filenames
                    # are PEP 440 compliant but not normalized.
                    # See: https://github.com/pypa/packaging/issues/527
                    and candidate.name == identifier
                ],
                key=attrgetter("version", "is_wheel"),
                reverse=True,
            )
        except PyPINotFoundError as e:
            yield SkippedCandidate(name=identifier, skip_reason=str(e))
            return

        logger.debug(f"{identifier} has candidates: {candidates}")

        # If we have multiple candidates for a single version and some are wheels,
        # yield only the wheels. This keeps us from wasting a large amount of
        # dependency search time when comparing wheels against source distributions.
        for _, candidates in itertools.groupby(candidates, key=attrgetter("version")):
            candidate = next(candidates)
            yield candidate
            if candidate.is_wheel:
                yield from (c for c in candidates if c.is_wheel)
            else:
                yield from candidates

    def is_satisfied_by(self, requirement: Any, candidate: Any) -> bool:
        """
        See `resolvelib.providers.AbstractProvider.is_satisfied_by`.
        """
        if isinstance(candidate, SkippedCandidate):
            return True

        # See the NOTE in find_matches: we use `filter(...)` because of its
        # special casing around prereleases.
        return any(
            requirement.specifier.filter(
                (candidate.version,),
            )
        )

    def get_dependencies(self, candidate: Any) -> Any:
        """
        See `resolvelib.providers.AbstractProvider.get_dependencies`.
        """
        if isinstance(candidate, SkippedCandidate):
            return []

        return candidate.dependencies


class PyPINotFoundError(Exception):
    """
    An error to signify that the provider could not find the requested project on PyPI.
    """

    pass



"""
Resolve a list of dependencies via the `resolvelib` API as well as a custom
`Resolver` that uses PyPI as an information source.
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import cast

from packaging.requirements import Requirement
from requests.exceptions import HTTPError
from resolvelib import BaseReporter, Resolver
from resolvelib.resolvers import ResolutionImpossible

from pip_audit._cache import caching_session
from pip_audit._dependency_source import (
    DependencyResolver,
    DependencyResolverError,
    HashMismatchError,
    HashMissingError,
    RequirementHashes,
)
from pip_audit._dependency_source.requirement import RequirementDependency
from pip_audit._service.interface import Dependency, SkippedDependency
from pip_audit._state import AuditState

from .pypi_provider import Candidate, PyPIProvider, ResolvedCandidate, SkippedCandidate

logger = logging.getLogger(__name__)

PYPI_URL = "https://pypi.org/simple/"


class ResolveLibResolver(DependencyResolver):
    """
    An implementation of `DependencyResolver` that uses `resolvelib` as its
    backend dependency resolution strategy.
    """

    def __init__(
        self,
        index_urls: list[str] = [PYPI_URL],
        timeout: int | None = None,
        cache_dir: Path | None = None,
        state: AuditState = AuditState(),
    ) -> None:
        """
        Create a new `ResolveLibResolver`.

        `timeout` and `cache_dir` are optional arguments for HTTP timeouts
        and caching, respectively.

        `state` is an `AuditState` to use for state callbacks.
        """
        self.index_urls = index_urls
        self.timeout = timeout
        # We keep the session here rather than create it within the provider. This is easier to mock
        # since we're creating a new provider on every `resolve` call.
        self.session = caching_session(cache_dir, use_pip=True)
        self.state = state
        self.reporter = BaseReporter()

    def resolve(self, reqs: list[Requirement], req_hashes: RequirementHashes) -> list[Dependency]:
        """
        Resolve the given `Requirement` into a `Dependency` list.
        """

        deps: list[Dependency] = []

        provider = PyPIProvider(self.index_urls, req_hashes, self.session, self.timeout, self.state)
        resolver: Resolver = Resolver(provider, self.reporter)

        try:
            result = resolver.resolve(reqs)
        except HTTPError as e:
            raise ResolveLibResolverError("failed to resolve dependencies") from e
        except ResolutionImpossible as e:
            raise ResolveLibResolverError(f"impossible resolution: {e}")

        # If the provider encountered any dependencies it couldn't find on PyPI, they'll be here.
        deps.extend(provider.skip_deps)

        # Construct dependee mapping to figure out what top-level requirements correspond to what
        # dependencies.
        dependee_map = _build_dependee_map(list(result.mapping.values()))

        for name, candidate in result.mapping.items():
            # Check whether the candidate was skipped
            if isinstance(candidate, SkippedCandidate):
                deps.append(SkippedDependency(candidate.name, candidate.skip_reason))
                continue

            # Check hash validity
            if req_hashes:
                try:
                    req_hashes.match(name, candidate.dist_hashes)
                except HashMissingError as e:
                    raise ResolveLibResolverError(
                        "Found dependency without an associated requirements "
                        f"file hash: {str(e)}"
                    ) from e
                except HashMismatchError as e:
                    raise ResolveLibResolverError(str(e)) from e

            dependee_reqs = _dependee_reqs_for_candidate(candidate, dependee_map, reqs)
            deps.append(RequirementDependency(name, candidate.version, dependee_reqs=dependee_reqs))
        return deps


class ResolveLibResolverError(DependencyResolverError):
    """
    A `resolvelib`-specific `DependencyResolverError`.
    """

    pass


def _build_dependee_map(
    candidates: list[Candidate],
) -> dict[Requirement, list[Requirement]]:
    """
    Build a mapping of dependee `Requirement`s to dependers. This is needed to find the top-level
    requirements that each subdependency originates from.
    """
    dependee_map: dict[Requirement, list[Requirement]] = {}
    for c in candidates:
        if isinstance(c, SkippedCandidate):
            continue
        c = cast(ResolvedCandidate, c)
        for dep in c.dependencies:
            if dep not in dependee_map:
                dependee_map[dep] = []
            dependee_map[dep].extend(c.reqs)
    return dependee_map


def _dependee_reqs_for_candidate(
    candidate: ResolvedCandidate,
    dependee_map: dict[Requirement, list[Requirement]],
    reqs: list[Requirement],
) -> set[Requirement]:
    """
    Find the set of top-level `Requirement`s that a given candidate originates from.
    """
    # Make sure we don't get stuck in a loop if the requirements file has cyclical dependencies.
    seen: list[Requirement] = list()

    def find_dependees(req: Requirement) -> set[Requirement]:
        if req in seen:
            return set()
        if req in reqs:
            return {req}
        seen.append(req)
        dependees = set()
        for r in dependee_map[req]:
            dependees |= find_dependees(r)
        seen.pop()
        return dependees

    dependee_reqs = set()
    for req in candidate.reqs:
        dependee_reqs |= find_dependees(req)
    return dependee_reqs



"""
Collect dependencies from one or more `requirements.txt`-formatted files.
"""

from __future__ import annotations

import logging
import os
import re
import shutil
from contextlib import ExitStack
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import IO, Iterator, cast

from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet
from packaging.version import Version
from pip_requirements_parser import InstallRequirement, InvalidRequirementLine, RequirementsFile

from pip_audit._dependency_source import (
    DependencyFixError,
    DependencyResolver,
    DependencyResolverError,
    DependencySource,
    DependencySourceError,
    RequirementHashes,
)
from pip_audit._fix import ResolvedFixVersion
from pip_audit._service import Dependency
from pip_audit._service.interface import ResolvedDependency, SkippedDependency
from pip_audit._state import AuditState

logger = logging.getLogger(__name__)

PINNED_SPECIFIER_RE = re.compile(r"==(?P<version>.+?)$", re.VERBOSE)


class RequirementSource(DependencySource):
    """
    Wraps `requirements.txt` dependency resolution as a dependency source.
    """

    def __init__(
        self,
        filenames: list[Path],
        resolver: DependencyResolver,
        *,
        require_hashes: bool = False,
        no_deps: bool = False,
        skip_editable: bool = False,
        state: AuditState = AuditState(),
    ) -> None:
        """
        Create a new `RequirementSource`.

        `filenames` provides the list of filepaths to parse.

        `resolver` is the `DependencyResolver` instance to use.

        `require_hashes` controls the hash policy: if `True`, dependency collection
        will fail unless all requirements include hashes.

        `no_deps` controls the dependency resolution policy: if `True`,
        dependency resolution is not performed and the inputs are checked
        and treated as "frozen".

        `skip_editable` controls whether requirements marked as "editable" are skipped.
        By default, editable requirements are not skipped.

        `state` is an `AuditState` to use for state callbacks.
        """
        self._filenames = filenames
        self._resolver = resolver
        self._require_hashes = require_hashes
        self._no_deps = no_deps
        self._skip_editable = skip_editable
        self.state = state
        self._dep_cache: dict[Path, set[Dependency]] = {}

    def collect(self) -> Iterator[Dependency]:
        """
        Collect all of the dependencies discovered by this `RequirementSource`.

        Raises a `RequirementSourceError` on any errors.
        """
        collected: dict[str, Dependency] = dict()
        for filename in self._filenames:
            try:
                rf = RequirementsFile.from_file(filename)
                if rf.invalid_lines:
                    raise RequirementSourceError(
                        f"requirement file {filename} contains invalid lines: "
                        f"{str(rf.invalid_lines)}"
                    )

                reqs: list[InstallRequirement] = []
                req_names: set[str] = set()
                for req in rf.requirements:
                    if req.req is None:
                        # For URL requirements that don't have an egg fragment that lists the
                        # package name and version, `pip-requirements-parser` won't attach a
                        # `Requirement` object to the `InstallRequirement`.
                        #
                        # In this case, we can't audit the dependency so we should signal to the
                        # caller that we're skipping it.
                        yield SkippedDependency(
                            name=req.requirement_line.line,
                            skip_reason="could not deduce package/specifier pair from requirement, "
                            "please specify them with #egg=your_package_name==your_package_version",
                        )
                        continue
                    if self._skip_editable and req.is_editable:
                        yield SkippedDependency(
                            name=req.name, skip_reason="requirement marked as editable"
                        )
                    if req.marker is None or req.marker.evaluate():
                        # This means we have a duplicate requirement for the same package
                        if req.name in req_names:
                            raise RequirementSourceError(
                                f"package {req.name} has duplicate requirements: {str(req)}"
                            )
                        req_names.add(req.name)
                        reqs.append(req)

                for dep in self._collect_cached_deps(filename, reqs):
                    if dep.canonical_name in collected:
                        existing_dep = collected[dep.canonical_name]
                        if isinstance(dep, SkippedDependency) or isinstance(
                            existing_dep, SkippedDependency
                        ):
                            # The `continue` statement is incorrectly flagged as uncovered for
                            # Python <= 3.9.
                            #
                            # Let's add a `pass` here as a way to make sure this branch gets tested
                            # and then mark the `continue` with `no cover`.
                            #
                            # See: https://github.com/pytest-dev/pytest-cov/issues/546
                            pass
                            continue  # pragma: no cover

                        dep = cast(RequirementDependency, dep)
                        existing_dep = cast(RequirementDependency, existing_dep)

                        # If we have the same dependency generated from multiple files, we need to
                        # merge the dependee requirements.
                        combined_dep = RequirementDependency(
                            name=dep.name,
                            version=dep.version,
                            dependee_reqs=(dep.dependee_reqs | existing_dep.dependee_reqs),
                        )

                        collected[dep.canonical_name] = combined_dep
                        continue

                    collected[dep.canonical_name] = dep
            except DependencyResolverError as dre:
                raise RequirementSourceError(str(dre))

        yield from collected.values()

    def fix(self, fix_version: ResolvedFixVersion) -> None:
        """
        Fixes a dependency version for this `RequirementSource`.
        """
        with ExitStack() as stack:
            # Make temporary copies of the existing requirements files. If anything goes wrong, we
            # want to copy them back into place and undo any partial application of the fix.
            tmp_files: list[IO[str]] = [
                stack.enter_context(NamedTemporaryFile(mode="w")) for _ in self._filenames
            ]
            for (filename, tmp_file) in zip(self._filenames, tmp_files):
                with filename.open("r") as f:
                    shutil.copyfileobj(f, tmp_file)

            try:
                # Now fix the files inplace
                for filename in self._filenames:
                    self.state.update_state(
                        f"Fixing dependency {fix_version.dep.name} ({fix_version.dep.version} => "
                        f"{fix_version.version})"
                    )
                    self._fix_file(filename, fix_version)
            except Exception as e:
                logger.warning(
                    f"encountered an exception while applying fixes, recovering original files: {e}"
                )
                self._recover_files(tmp_files)
                raise e

    def _fix_file(self, filename: Path, fix_version: ResolvedFixVersion) -> None:
        # Reparse the requirements file. We want to rewrite each line to the new requirements file
        # and only modify the lines that we're fixing.
        #
        # This time we're using the `RequirementsFile.parse` API instead of `Requirements.from_file`
        # since we want to access each line sequentially in order to rewrite the file.
        reqs = list(RequirementsFile.parse(filename=str(filename)))

        # Check ahead of time for anything invalid in the requirements file since we don't want to
        # encounter this while writing out the file. Check for duplicate requirements and lines that
        # failed to parse.
        req_names: set[str] = set()
        for req in reqs:
            if (
                isinstance(req, InstallRequirement)
                and (req.marker is None or req.marker.evaluate())
                and req.req is not None
            ):
                if req.name in req_names:
                    raise RequirementFixError(
                        f"package {req.name} has duplicate requirements: {str(req)}"
                    )
                req_names.add(req.name)
            elif isinstance(req, InvalidRequirementLine):
                raise RequirementFixError(
                    f"requirement file {filename} has invalid requirement: {str(req)}"
                )

        # Now write out the new requirements file
        with filename.open("w") as f:
            fixed = False
            for req in reqs:
                if (
                    isinstance(req, InstallRequirement)
                    and req.name == fix_version.dep.name
                    and req.specifier.contains(fix_version.dep.version)
                    and not req.specifier.contains(fix_version.version)
                ):
                    req.req.specifier = SpecifierSet(f"=={fix_version.version}")
                    fixed = True
                print(req.dumps(), file=f)

            # The vulnerable dependency may not be explicitly listed in the requirements file if it
            # is a subdependency of a requirement. In this case, we should explicitly add the fixed
            # dependency into the requirements file.
            #
            # To know whether this is the case, we'll need to resolve dependencies if we haven't
            # already in order to figure out whether this subdependency belongs to this file or
            # another.
            if not fixed:
                req_dep = cast(RequirementDependency, fix_version.dep)
                if req_dep.dependee_reqs:
                    logger.warning(
                        "added fixed subdependency explicitly to requirements file "
                        f"{filename}: {fix_version.dep.canonical_name}"
                    )
                    dependee_reqs_formatted = ",".join(
                        [
                            str(req)
                            for req in sorted(list(req_dep.dependee_reqs), key=lambda x: x.name)
                        ]
                    )
                    print(
                        f"    # pip-audit: subdependency fixed via {dependee_reqs_formatted}",
                        file=f,
                    )
                    print(f"{fix_version.dep.canonical_name}=={fix_version.version}", file=f)

    def _recover_files(self, tmp_files: list[IO[str]]) -> None:
        for (filename, tmp_file) in zip(self._filenames, tmp_files):
            try:
                os.replace(tmp_file.name, filename)
                # We need to tinker with the internals to prevent the file wrapper from attempting
                # to remove the temporary file like in the regular case.
                tmp_file._closer.delete = False  # type: ignore[attr-defined]
            except Exception as e:
                # Not much we can do at this point since we're already handling an exception. Just
                # log the error and try to recover the rest of the files.
                logger.warning(f"encountered an exception during file recovery: {e}")
                continue

    def _collect_preresolved_deps(
        self,
        reqs: Iterator[InstallRequirement],
    ) -> Iterator[tuple[Requirement, Dependency]]:
        """
        Collect pre-resolved (pinned) dependencies.
        """
        for req in reqs:
            # NOTE: URL dependencies cannot be pinned, so skipping them
            # makes sense (under the same principle of skipping dependencies
            # that can't be found on PyPI). This is also consistent with
            # what `pip --no-deps` does (installs the URL dependency, but
            # not any subdependencies).
            if req.is_url:
                yield req.req, SkippedDependency(
                    name=req.name,
                    skip_reason="URL requirements cannot be pinned to a specific package version",
                )
            elif not req.specifier:
                raise RequirementSourceError(f"requirement {req.name} is not pinned: {str(req)}")
            else:
                pinned_specifier = PINNED_SPECIFIER_RE.match(str(req.specifier))
                if pinned_specifier is None:
                    raise RequirementSourceError(
                        f"requirement {req.name} is not pinned to an exact version: {str(req)}"
                    )

                yield req.req, RequirementDependency(
                    req.name, Version(pinned_specifier.group("version"))
                )

    def _build_hash_options_mapping(self, hash_options: list[str]) -> dict[str, list[str]]:
        """
        A helper that takes a list of hash options and returns a dictionary mapping from hash
        algorithm (e.g. sha256) to a list of values.
        """
        mapping: dict[str, list[str]] = {}
        for hash_option in hash_options:
            algorithm, hash_ = hash_option.split(":")
            if algorithm not in mapping:
                mapping[algorithm] = []
            mapping[algorithm].append(hash_)
        return mapping

    def _collect_cached_deps(
        self, filename: Path, reqs: list[InstallRequirement]
    ) -> Iterator[Dependency]:
        """
        Collect resolved dependencies for a given requirements file, retrieving them from the
        dependency cache if possible.
        """
        # See if we've already have cached dependencies for this file
        cached_deps_for_file = self._dep_cache.get(filename, None)
        if cached_deps_for_file is not None:
            yield from cached_deps_for_file

        new_cached_deps_for_file: set[Dependency] = set()

        # Skip dependency resolution if the user has specified `--no-deps`
        if self._no_deps:
            for req, dep in self._collect_preresolved_deps(iter(reqs)):
                new_cached_deps_for_file.add(dep)
                yield dep
        else:
            require_hashes = self._require_hashes or any(req.hash_options for req in reqs)
            req_hashes = RequirementHashes()

            # If we're requiring hashes, enforce that all requirements are hashed
            if require_hashes:
                for hash_req in reqs:
                    if not hash_req.hash_options:
                        raise RequirementSourceError(
                            f"requirement {hash_req.name} does not contain a hash {str(hash_req)}"
                        )
                    req_hashes.add_req(
                        hash_req.name, self._build_hash_options_mapping(hash_req.hash_options)
                    )

            # Invoke the dependency resolver to turn requirements into dependencies
            req_values: list[Requirement] = [r.req for r in reqs]
            for dep in self._resolver.resolve(req_values, req_hashes):
                new_cached_deps_for_file.add(dep)

                if dep.is_skipped():  # pragma: no cover
                    dep = cast(SkippedDependency, dep)
                    self.state.update_state(f"Skipping {dep.name}: {dep.skip_reason}")
                else:
                    dep = cast(ResolvedDependency, dep)
                    self.state.update_state(f"Collecting {dep.name} ({dep.version})")

                yield dep

        # Cache the collected dependencies
        self._dep_cache[filename] = new_cached_deps_for_file


class RequirementSourceError(DependencySourceError):
    """A requirements-parsing specific `DependencySourceError`."""

    pass


class RequirementFixError(DependencyFixError):
    """A requirements-fixing specific `DependencyFixError`."""

    pass


@dataclass(frozen=True)
class RequirementDependency(ResolvedDependency):
    """
    Represents a fully resolved Python package from a requirements file.
    """

    dependee_reqs: set[Requirement] = field(default_factory=set, hash=False)
